/*
 * Copyright 2005-2018 ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 *
 * In applying this licence, ECMWF does not waive the privileges and immunities granted to it by
 * virtue of its status as an intergovernmental organisation nor does it submit to any jurisdiction.
 */

/**********************************
 *  Enrico Fucile
 **********************************/

#include "grib_api_internal.h"
/*
   This is used by make_class.pl

   START_CLASS_DEF
   CLASS      = accessor
   SUPER      = grib_accessor_class_long
   IMPLEMENTS = unpack_long
   IMPLEMENTS = init
   MEMBERS = const char* values
   MEMBERS = const char* bitsPerValue
   MEMBERS = const char* numberOfPoints
   MEMBERS = const char* bitmapPresent
   MEMBERS = const char* bitmap
   MEMBERS = const char* numberOfCodedValues
   END_CLASS_DEF
 */

/* START_CLASS_IMP */

/*

Don't edit anything between START_CLASS_IMP and END_CLASS_IMP
Instead edit values between START_CLASS_DEF and END_CLASS_DEF
or edit "accessor.class" and rerun ./make_class.pl

*/

static int unpack_long(grib_accessor*, long* val,size_t *len);
static void init(grib_accessor*,const long, grib_arguments* );
static void init_class(grib_accessor_class*);

typedef struct grib_accessor_number_of_values {
    grib_accessor          att;
/* Members defined in gen */
/* Members defined in long */
/* Members defined in number_of_values */
	const char* values;
	const char* bitsPerValue;
	const char* numberOfPoints;
	const char* bitmapPresent;
	const char* bitmap;
	const char* numberOfCodedValues;
} grib_accessor_number_of_values;

extern grib_accessor_class* grib_accessor_class_long;

static grib_accessor_class _grib_accessor_class_number_of_values = {
    &grib_accessor_class_long,                      /* super                     */
    "number_of_values",                      /* name                      */
    sizeof(grib_accessor_number_of_values),  /* size                      */
    0,                           /* inited */
    &init_class,                 /* init_class */
    &init,                       /* init                      */
    0,                  /* post_init                      */
    0,                    /* free mem                       */
    0,                       /* describes himself         */
    0,                /* get length of section     */
    0,              /* get length of string      */
    0,                /* get number of values      */
    0,                 /* get number of bytes      */
    0,                /* get offset to bytes           */
    0,            /* get native type               */
    0,                /* get sub_section                */
    0,               /* grib_pack procedures long      */
    0,               /* grib_pack procedures long      */
    0,                  /* grib_pack procedures long      */
    &unpack_long,                /* grib_unpack procedures long    */
    0,                /* grib_pack procedures double    */
    0,              /* grib_unpack procedures double  */
    0,                /* grib_pack procedures string    */
    0,              /* grib_unpack procedures string  */
    0,                 /* grib_pack procedures bytes     */
    0,               /* grib_unpack procedures bytes   */
    0,            /* pack_expression */
    0,              /* notify_change   */
    0,                /* update_size   */
    0,            /* preferred_size   */
    0,                    /* resize   */
    0,      /* nearest_smaller_value */
    0,                       /* next accessor    */
    0,                    /* compare vs. another accessor   */
    0,     /* unpack only ith value          */
    0,     /* unpack a subarray         */
    0,             		/* clear          */
};


grib_accessor_class* grib_accessor_class_number_of_values = &_grib_accessor_class_number_of_values;


static void init_class(grib_accessor_class* c)
{
	c->dump	=	(*(c->super))->dump;
	c->next_offset	=	(*(c->super))->next_offset;
	c->string_length	=	(*(c->super))->string_length;
	c->value_count	=	(*(c->super))->value_count;
	c->byte_count	=	(*(c->super))->byte_count;
	c->byte_offset	=	(*(c->super))->byte_offset;
	c->get_native_type	=	(*(c->super))->get_native_type;
	c->sub_section	=	(*(c->super))->sub_section;
	c->pack_missing	=	(*(c->super))->pack_missing;
	c->is_missing	=	(*(c->super))->is_missing;
	c->pack_long	=	(*(c->super))->pack_long;
	c->pack_double	=	(*(c->super))->pack_double;
	c->unpack_double	=	(*(c->super))->unpack_double;
	c->pack_string	=	(*(c->super))->pack_string;
	c->unpack_string	=	(*(c->super))->unpack_string;
	c->pack_bytes	=	(*(c->super))->pack_bytes;
	c->unpack_bytes	=	(*(c->super))->unpack_bytes;
	c->pack_expression	=	(*(c->super))->pack_expression;
	c->notify_change	=	(*(c->super))->notify_change;
	c->update_size	=	(*(c->super))->update_size;
	c->preferred_size	=	(*(c->super))->preferred_size;
	c->resize	=	(*(c->super))->resize;
	c->nearest_smaller_value	=	(*(c->super))->nearest_smaller_value;
	c->next	=	(*(c->super))->next;
	c->compare	=	(*(c->super))->compare;
	c->unpack_double_element	=	(*(c->super))->unpack_double_element;
	c->unpack_double_subarray	=	(*(c->super))->unpack_double_subarray;
	c->clear	=	(*(c->super))->clear;
}

/* END_CLASS_IMP */

static void init(grib_accessor* a,const long l, grib_arguments* c)
{
  int n=0;
  grib_accessor_number_of_values* self = (grib_accessor_number_of_values*)a;
  self->values = grib_arguments_get_name(a->parent->h,c,n++);
  self->bitsPerValue = grib_arguments_get_name(a->parent->h,c,n++);
  self->numberOfPoints = grib_arguments_get_name(a->parent->h,c,n++);
  self->bitmapPresent = grib_arguments_get_name(a->parent->h,c,n++);
  self->bitmap = grib_arguments_get_name(a->parent->h,c,n++);
  self->numberOfCodedValues = grib_arguments_get_name(a->parent->h,c,n++);
  a->flags  |= GRIB_ACCESSOR_FLAG_READ_ONLY;
  
  a->length=0;
}

static int  unpack_long(grib_accessor* a, long* val, size_t *len)
{
  grib_accessor_number_of_values* self = (grib_accessor_number_of_values*)a;
  int ret=GRIB_SUCCESS,i;
  long npoints=0,bitmap_present=0;
  size_t size=0;

#if GRIB_WARNING
  long numberOfCodedValues=0;
  long bpv=0;
  if((ret = grib_get_long_internal(a->parent->h, self->bitsPerValue,&bpv)) != GRIB_SUCCESS)
    return ret;
#endif

  if((ret = grib_get_long_internal(a->parent->h, self->numberOfPoints,&npoints)) != GRIB_SUCCESS)
      return ret;

  if((ret = grib_get_long_internal(a->parent->h, self->bitmapPresent,&bitmap_present)) != GRIB_SUCCESS)
      return ret;

  if (bitmap_present) {
      double* bitmap;
      size=npoints;
      bitmap=(double*)grib_context_malloc(a->parent->h->context,sizeof(double)*size);
      if((ret = grib_get_double_array_internal(a->parent->h,self->bitmap,bitmap,&size))
                  != GRIB_SUCCESS) {
			  grib_context_free(a->parent->h->context,bitmap);
			  return ret;
	  }
      *val=0;
      for (i=0;i<size;i++)
        if (bitmap[i]!=0) (*val)++;

      grib_context_free(a->parent->h->context,bitmap);
  } else {
      *val=npoints;
  }

#if GRIB_WARNING
  if ( bpv != 0 ) {
    if((ret = grib_get_long_internal(a->parent->h, self->numberOfCodedValues,&numberOfCodedValues)) != GRIB_SUCCESS)
      return ret;
     if (*val != numberOfCodedValues) {
       grib_context_log(a->parent->h->context,GRIB_LOG_WARNING,
       "number of values from data section (%ld) doesn't match geography (%ld)",numberOfCodedValues,*val);
       /* *val=numberOfCodedValues; */
  	}
			   
  }
#endif


  return ret;
}

